""" Main flask app module """

from flask import Flask, request, jsonify, send_file, redirect, url_for
from flask_login import LoginManager, login_user, logout_user, login_required, current_user
from flask_bootstrap import Bootstrap
from flask_restful import Api, Resource
from flasgger import Swagger
from flask_cors import CORS
from functools import wraps
import logging
from .extensions import infras, initialize, users
"""
from app.infra import Infra, Infras, initialize
from app.model import User, Users

initialize() # data files are saved in infra folder, html/js index file in web, API in api folder
infras = Infras.load_from_file("data/infras.json")
"""

VALID_API_KEYS = {"aj_pogodi", "another_test_api_key"}

"""
def require_api_key(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        api_key = request.headers.get('x-api-key')
        if api_key and api_key in VALID_API_KEYS:
            return f(*args, **kwargs)
        else:
            return jsonify({"error": "Unauthorized access"}), 401
    return decorated_function
"""

def require_api_key(f): # XXX: change back to api keys asap
    @wraps(f)
    def decorated_function(*args, **kwargs):
        api_key = request.headers.get('x-api-key')
        return f(*args, **kwargs)
    return decorated_function


app = Flask(__name__)
# use bootstrap for reactive apps, make it use local files and NOT from CDN
bootstrap = Bootstrap(app)
app.config['BOOTSTRAP_SERVE_LOCAL'] = True

PREFIX = 'store/'
app.config['STORE_FOLDER'] = 'store'

STATIC_FOLDER = 'static'
app.config['STATIC_FOLDER'] = STATIC_FOLDER

TEMP_FOLDER = PREFIX+'temp'
app.config['TEMP_FOLDER'] = TEMP_FOLDER

app.config['MAX_USERS'] = 20

#SECRET_KEY = os.urandom(32)
SECRET_KEY = 'abrakadabra'
app.config['SECRET_KEY'] = SECRET_KEY

# create Blueprints
from .blueprints.api import bp as api_bp # get api bp
app.register_blueprint(api_bp, url_prefix='/api')
from .blueprints.auth import bp as auth_bp # get auth bp
app.register_blueprint(auth_bp, url_prefix='/auth')
from .blueprints.users import bp as users_bp # get users bp
app.register_blueprint(users_bp, url_prefix='/users')
from .blueprints.web import bp as web_bp # get web bp
app.register_blueprint(web_bp, url_prefix='/web')

from flask_wtf.csrf import CSRFProtect
csrf = CSRFProtect(app) #CSRF web bp

from flask_wtf.csrf import CSRFError
@web_bp.errorhandler(CSRFError)
def handle_csrf_error(e):
    logging.warning('CSRF validation failed: %s',e.description)
    flash('Please login. Session expited and CSRF validation failed: ' + e.description,'warnings')
    # del users[current_user.get_id()] # rm user from cached 
    logout_user()   # Tell flask-login to log him out.
    return redirect(url_for('login'))                

csrf.exempt(api_bp) # exempt REST api from CSRF
csrf.exempt(auth_bp) # exempt REST auth from CSRF
csrf.exempt(users_bp) # exempt REST users from CSRF

# Initialize Flask-Login
login_manager = LoginManager()
login_manager.init_app(app)

@login_manager.user_loader
def load_user(username):
    u = users.get_user(username)
    return u

@login_manager.unauthorized_handler
def unauthorized_callback():
    logging.warning('request.path %s ',request.path)
    return redirect(url_for('web.login')+'?next=' + request.path)

#logging.basicConfig(level=logging.ERROR)
#logging.basicConfig(format='%(asctime)s - %(message)s', datefmt='%d-%b-%y %H:%M:%S')

api = Api(app)

swagger = Swagger(app, template={
    "info": {
        "title": "Infra API",
        "description": "MVP server and data model",
        "version": "0.0.1"
    }
})

@app.route('/', methods=['GET'])
def index():
    return send_file('static/index.html')

@app.route('/favicon.ico')
def favicon():
    #return send_from_directory(app.config['STATIC_FOLDER'], 'favicon.ico',mimetype='image/vnd.microsoft.icon')
    #return send_from_directory(app.config['STATIC_FOLDER'], 'favicon.png')
    return send_file(app.config['STATIC_FOLDER']+'/favicon.png')

if __name__ == '__main__':
    app.run(debug=True)
