'''

api blueprint, using infra module

function is to make API for infra CRUD access

'''
#import json
from flask import request, jsonify, make_response, json
from flask import Flask, jsonify, request 
from flask import Blueprint
from flasgger import swag_from
from app import infras, require_api_key
from app.extensions import infras

bp = Blueprint('api',__name__, url_prefix='/api')

# xjsonify because jsonify incorrectly serializes my objects
def xjsonify(x):
    response = make_response( x.to_json() )
    response.headers["Content-Type"] = "application/json"
    return response

@bp.errorhandler(404)
def not_found(error):
    return jsonify({"message": "Endpoint not found"}), 404


# get list of all target entities=(infras) or error if not exist
@bp.route('/<target>', methods = ['GET'])
@require_api_key
def list(target=None):
    """Example endpoint returning a list of target entities
    ---
    tags:
        - List of saved target entities in the database
    description: Returns list of ids for the saved target entities
    parameters:
      - name: target
        in: path
        type: string
        required: id
        default: 'entities'
    responses:
      200:
        description: An JSON list of infra =ids (that may later be addresed by these ids)
        type: array
        items:
          type: string
        examples:
          application/json:
            - entities
            - infras
      404:
        description: Asked for a wrong entity
        examples:
          application/json:
            messg:"target entity foo not found"
    """    
    if(request.method == 'GET'):
        t = None
        if (target is None) or (target==''):
            target = 'entities'
        if (target not in ('entities', 'infras')):
            return jsonify({'messg':f'target entity {target} not found'}) , 404 
        if target == 'entities':
            #print("Handling entities")
            t = ['entities', 'infras']
            return jsonify(t)#. 200 this is hardly a JSON, this is a list
        elif target == 'infras':
            t = infras
        else:
            #print("Unknown option")
            return jsonify({'messg':f'target entity {target} not found. how on earth you made this error?'}) , 404 
        return jsonify(t.ls())#. 200 this is hardly a JSON, this is a list

# --- infra ==============================================================================
# get a infra if id!=None 
# delete a infra for given id, if exist or error if not exist
@bp.route('/infras/<id>', methods = ['GET', 'DELETE'])
#@swag_from('blueprints/infra.yml')
@require_api_key
def infra_get_del(id=None):
    """Endpoint returning a infra by id, or deleting a infra by id
    ---
    tags:
        - infras manipulation. Get ids from list endpoint
    parameters:
      - name: id
        in: path
        type: string
        required: true
        default: none
    definitions:
      Infra:
        type: object
        properties:
          id:
            type: string
            description: "Unique id"
          description:
            type: string
            description: "Common description"
          attribute_1:
            type: string
            description: "Attribute 1"
          attribute_2:
            type: string
          boxes:
            type: object
          metrics_boxes:
            type: object
          is_paused:
            type: string
            description: "Boolean if infra is paused or not"
          required:
            - id
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
    responses:
      200:
        description: An JSON infra object addresed by id string
        schema:
          $ref: '#/definitions/Infra'
        examples:
          application/json:
            [   "Assets",  "Security Attributes",  "Security Controls",  "/",  "/home"]
      204:
        description: A JSON infra object deleted
        schema:
          $ref: '#/definitions/ErrMsg'
        examples:
          application/json:
            messg:'A JSON infra object deleted'
      404:
        description: A JSON infra object not found error message
        schema:
          $ref: '#/definitions/ErrMsg'
        examples:
          application/json:
            messg:'infra with this id not found'
    """    
    if(request.method == 'GET'): 
        infra = None
        try:
            infra = infras.get(id)
        except ValueError as ve:
            return jsonify({'messg':f'infra with this id={id} not found'}) , 404 
        if infra is not None:
            return xjsonify(infra), 200
        else:
            return jsonify({'messg':f'infra with this id={id} not found'}) , 404 
    if(request.method == 'DELETE'): 
        infra = None
        try:
            infra = infras.get(id)
        except ValueError as ve:
            return jsonify({'messg':f'infra with this id={id} not found'}) , 404 
        if infra is not None:
            infras.remove(infra)
            # check if removed
            ds = infras.get(id)
            if ds is not None:
                return jsonify({'messg':f'infra with this id={id} could not be removed'}) , 400
            else:
                return jsonify({'messg':f'infra with this id={id} is removed'}) , 204 
        else:
            return jsonify({'messg':f'infra with this id={id} not found'}) , 404 


# post a infra if id!=None and a infra is passed in the request body
@bp.route('/infras', methods = ['POST'])
#@swag_from('blueprints/infra.yml')
@require_api_key
def infra_post():
    """Endpoint saving a infra. id and infra object in request body need to have a matching id
    ---
    tags:
        - infras manipulation. Get ids from list endpoint
    parameters:
      - name: a_infra
        in: body
        schema: 
          $ref: '#/definitions/Infra'
        required: true
    definitions:
      Infra:
        type: object
        properties:
          id:
            type: string
            description: "Unique id"
          description:
            type: string
            description: "Common description"
          attribute_1:
            type: string
            description: "Attribute 1"
          attribute_2:
            type: string
          boxes:
            type: object
          metrics_boxes:
            type: object
          is_paused:
            type: string
            description: "Boolean if infra is paused or not"
          required:
            - id
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
    responses:
      201:
        description: An JSON infra object saved
        schema:
          $ref: '#/definitions/ErrMsg'
        examples:
          application/json:
            messg:'An JSON infra object saved'
      409:
        description: A JSON infra object already exist
        schema:
          $ref: '#/definitions/ErrMsg'
        examples:
          application/json:
            messg:'A JSON infra object already exist'
      422:
        description: A JSON infra objects not deserializable, check your JSON. 
        schema:
          $ref: '#/definitions/ErrMsg'
        examples:
          application/json:
            messg:'A JSON infra objects not deserializable, check your JSON. '
    """    
    if(request.method == 'POST'): 
        infra = None
        try:
            infra = infra.from_json(request.data)
        except ValueError as e:
            return jsonify({'messg':f'infra not deserializable, check your JSON. Caught an exception: {e}'}) , 422
        except KeyError as ke:
            return jsonify({'messg':f'infra not deserializable, check your JSON. Caught an exception: {ke}'}) , 422        
        if infra is not None: # have the object
            if infras.add(infra): # save in infras list
                # first save new infra values to back end, then return
                return jsonify({'messg':f'infra id={infra.id} saved'}) , 201
            else:
                return jsonify({'messg':f'infra with this id={infra.id} already exist'}) , 409
        else:
            return jsonify({'messg':'infra not deserializable, check your JSON'}) , 422


