'''
api authentication blueprint, using users module
'''
import jwt
from datetime import datetime, timedelta
from flask import Blueprint, jsonify, request
from flask_login import current_user, login_user
from werkzeug.security import safe_str_cmp
from app import users, infras, require_api_key
from app.extensions import hash_me


# JWT secret and settings
JWT_SECRET = "your-secret-key"  # Replace with a secure secret key
JWT_ALGORITHM = "HS256"
JWT_EXP_DELTA_SECONDS = 3600  # Token expiration time in seconds

#from .api import bp
bp = Blueprint('auth',__name__, url_prefix='/auth')


"""
    Generate a JWT token for the authenticated user.
    Requires a username and password in the request JSON.
    curl -X POST http://localhost:5000/api/token -H "Content-Type: application/json" -d '{"username": "bob", "password": "password2"}'
"""
@bp.route("/token", methods=["POST"])
def token():
    """Endpoint for JWT token generation for the authenticated user.
    ---
    tags: 
        - Authentication API
    description: Endpoint for JWT token generation for the authenticated user.
    parameters:
      - name: body
        in: body
        schema:
          $ref: "#/definitions/JWTCredentials"
        required: true
    responses:
      200:
        description: "token returned"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            "token": "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJ1c2VybmFtZSI6ImJvYiIsImV4cCI6MTczNDk4OTM0NX0.yMoq_80WTZEtG8OlLGO1dWx22aEiVf_BA6mjR1uSEOg"
      400:
        description: "Username or password missing in supplied credentials"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            "error": "Username and password required"
      401:
        description: "Invalid credentials"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            "error": "Invalid credentials"
    definitions:
      JWTCredentials:
        type: object
        properties:
          username:
            type: string
            description: existing username
          password:
            type: string
            description: password
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
    """
    
    data = request.json
    if not data or "username" not in data or "password" not in data:
        return jsonify({"error": "Username and password required"}), 400
    # Authenticate user
    username = data["username"]
    password = data["password"]
    #user = next((u for u in users.values() if u["username"] == username), None)
    user = users.get_user(username)
    if user and safe_str_cmp(user.password, hash_me(password) ):
        # Generate JWT token
        payload = {
            "username": username,
            "exp": datetime.utcnow() + timedelta(seconds=JWT_EXP_DELTA_SECONDS)
        }
        token = jwt.encode(payload, JWT_SECRET, algorithm=JWT_ALGORITHM)
        return jsonify({"token": token})
    return jsonify({"error": "Invalid credentials"}), 401



from functools import wraps
# JWT settings
#JWT_SECRET = "your-secret-key"  # Replace with a secure secret key
#JWT_ALGORITHM = "HS256"
def jwt_required(f):
    """
    Decorator to protect routes with JWT authentication.
    Ensures the request has a valid token in the Authorization header.
    """
    @wraps(f)
    def decorated_function(*args, **kwargs):
        auth_header = request.headers.get("Authorization")
        if not auth_header:
            return jsonify({"error": "Missing Authorization header"}), 401
        try:
            # Extract and decode the JWT
            token = auth_header.split("Bearer ")[1]
            decoded = jwt.decode(token, JWT_SECRET, algorithms=[JWT_ALGORITHM])
            # Attach decoded payload to the request context (optional)
            request.user = decoded  # e.g., decoded['user_id'], decoded['username']
        except jwt.ExpiredSignatureError:
            return jsonify({"error": "Token expired"}), 401
        except jwt.InvalidTokenError:
            return jsonify({"error": "Invalid token"}), 401
        # Proceed to the original route function
        return f(*args, **kwargs)
    return decorated_function

"""
    Example protected API route using the jwt_required decorator.
    curl -X GET http://localhost:5000/api/test -H "Authorization: Bearer <your-valid-token>"
    curl -X GET http://localhost:5000/api/test
    curl -X GET http://localhost:5000/api/test -H "Authorization: Bearer invalid-token"
"""
@bp.route("/test", methods=["GET", "POST"])
@jwt_required
def api_data():
    """Endpoint for JWT token testing of the authenticated user.
    ---
    tags: 
        - Authentication API
    description: Endpoint for JWT token testing of the authenticated user.
    parameters:
      - name: Authorization
        in: header
        type: string
        required: true
        description: |
          Bearer token for authentication. Format: `Bearer <your-valid-token>`
    responses:
      200:
        description: "Token returned"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            "message": "Access granted for user bob."
      401:
        description: "Invalid credentials"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            "error": "Invalid credentials"
    definitions:
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
            description: "Response message indicating success or error"
    """

    # Access the decoded user information from the request context
    user_info = request.user  # e.g., {'user_id': '2', 'username': 'api_user'}
    return jsonify({"message": f"Access granted for user {user_info['username']}."})

