'''

users blueprint, using user module

function is to make users for user CRUD access

'''
import json
from flask import request, jsonify, make_response, json
from flask import Flask, jsonify, request 
from flask import Blueprint
from flasgger import swag_from
from app.extensions import users as users_datastore
from app.extensions import User
from .auth import jwt_required

bp = Blueprint('users',__name__, url_prefix='/users')

# xjsonify because jsonify incorrectly serializes my objects
def xjsonify(x):
    response = make_response( x.to_json() )
    response.headers["Content-Type"] = "application/json"
    return response

@bp.errorhandler(404)
def not_found(error):
    return jsonify({"message": "Endpoint not found"}), 404


def remove_password(json_obj):
    """
    Removes the 'password' key from the JSON object if it exists.

    :param json_obj: dict, JSON object from which the password needs to be removed
    :return: dict, JSON object without the password key
    """
    if 'password' in json_obj:
        json_obj['password'] = '*****'
    return json_obj


# get list of all users or error if not exist
# /users GET full list of users if allowed
# /users/?=partial_username GET searched list of users if allowed
# curl -X GET "http://example.com/users?username=johndoe"
@bp.route('', methods=['GET'])
@jwt_required
def list():
    """Endpoint returning a list of usernames or filtering by username
    ---
    tags:
        - User directory
    description: Returns list of usernames for the saved users. Allows searching for a specific username.
    parameters:
      - name: username
        in: query
        type: string
        description: The username to search for
        required: false
    responses:
      200:
        description: A JSON list of usernames or the specific username if found
        type: array
        items:
          type: string
        examples:
          application/json:
            - usernames
      401:
        description: UNAUTHORIZED
        examples:
          application/json:
            "error": "Missing Authorization header"
      404:
        description: No users found or the specific username not found
        examples:
          application/json:
            messg: "username johndoe not found"
    """
    username = request.args.get('username')  # Get the 'username' query parameter
    if username:
        # Filter for the specific username
        t = [u.username for u in users_datastore.get_users() if username in u.username ]
        if not t:
            return jsonify({"messg": f"target entity {username} not found"}), 404
        return jsonify(t), 200
    else:
        # Return all usernames
        t = [u.username for u in users_datastore.get_users()]
        if not t:
            return jsonify({"messg": "No users found"}), 404
        return jsonify(t), 200

# get a user
# /users/<username> GET {username} or 404 if allowed
@bp.route('/<username>', methods = ['GET'])
@jwt_required
def get_user(username=None):
    """Endpoint returning user data
    ---
    tags: 
        - User directory
    description: Get a user data by username
    parameters:
      - name: username
        in: path
        type: string
        description: Username to get
        required: true
    definitions:
      User:
        type: object
        properties:
          username:
            type: string
            description: unique username
          gecos:
            type: string
            description: common name and family name
          role:
            type: string
            description: assigned role
          password:
            type: string
            description: will be masked
          email:
            type: string
          phone:
            type: string
          IM:
            type: string
          isLocked:
            type: boolean
            description: Boolean if the user is locked or not
        required:
          - username
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
    responses:
      200:
        description: "A JSON user object addressed by username string"
        schema:
            $ref: "#/definitions/User"
        examples:
          application/json:
            {
              "username": "johndoe",
              "gecos": "John Doe",
              "role": "admin",
              "password": "*****",
              "email": "johndoe@example.com",
              "phone": "+1234567890",
              "IM": "johndoe123",
              "isLocked": false
            }
      401:
        description: UNAUTHORIZED
        examples:
          application/json:
            "error": "Missing Authorization header"
      404:
        description: "A JSON user object not found error message"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "user with this username not found"
    definitions:
      User:
        type: object
        properties:
          username:
            type: string
            description: unique username
          gecos:
            type: string
            description: common name and family name
          role:
            type: string
            description: assigned role
          password:
            type: string
          email:
            type: string
          phone:
            type: string
          IM:
            type: string
          isLocked:
            type: boolean
            description: Boolean if the user is locked or not
        required:
          - username
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
    """
    if(request.method == 'GET'): 
        user = None
        try:
            user = users_datastore.get_user(username)
        except ValueError as ve:
            return jsonify({'messg':f'user with this username={username} not found'}) , 404 
        if user is not None:
            return remove_password(json.loads(user.to_json())), 200
            #return user.to_json(), 200
        else:
            return jsonify({'messg':f'user with this username={username} not found'}) , 404 

# delete a user
# /users/<username> DELETE {username} or 404 if allowed
@bp.route('/<username>', methods = ['DELETE'])
@jwt_required
def del_users(username=None):
    """Endpoint deleting user data
    ---
    tags: 
        - User directory
    description: Endpoint deleting user data
    parameters:
      - name: username
        in: path
        type: string
        description: Username to delete
        required: true
    responses:
      204:
        description: "User deleted"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "User deleted"
      401:
        description: UNAUTHORIZED
        examples:
          application/json:
            "error": "Missing Authorization header"
      404:
        description: "User not found error message"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "user with this username not found"
    definitions:
      User:
        type: object
        properties:
          username:
            type: string
            description: unique username
          gecos:
            type: string
            description: common name and family name
          role:
            type: string
            description: assigned role
          password:
            type: string
          email:
            type: string
          phone:
            type: string
          IM:
            type: string
          isLocked:
            type: boolean
            description: Boolean if the user is locked or not
        required:
          - username
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
    """
    if(request.method == 'DELETE'): 
        user = None
        try:
            user = users_datastore.get_user(username)
        except ValueError as ve:
            return jsonify({'messg':f'user with this username={username} not found'}) , 404 
        if user is not None:
            # users_datastore.remove(user), if my role allows
            my_username = request.user['username'] # request.user is decoded from jwt token
            if my_username == username: # cannot delete self
                                return jsonify({'messg':f'user cannot delete self'}) , 400
            my_role = users_datastore.get_my_role(my_username) # fetch my role
            users_datastore.del_user(my_role, username)
            # check if removed
            ds = users_datastore.get_user(username)
            if ds is not None:
                return jsonify({'messg':f'user with this username={username} could not be removed'}) , 400
            else:
                return jsonify({'messg':f'user with this username={username} is removed'}) , 204 
        else:
            return jsonify({'messg':f'user with this username={username} not found'}) , 404 


# post a user if username!=None and a user is passed in the request body
@bp.route('/<username>', methods = ['POST'])
@jwt_required
def user_post(username=None):
    """Endpoint to POST new user data
    ---
    tags: 
        - User directory
    description: Endpoint to save a new user data
    parameters:
      - name: username
        in: path
        type: string
        required: true
      - name: body
        in: body
        schema:
          $ref: "#/definitions/User"
        required: true
    responses:
      201:
        description: "User object saved"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "User object saved"
      401:
        description: UNAUTHORIZED
        examples:
          application/json:
            "error": "Missing Authorization header"
      409:
        description: "User object already exists"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "User object already exists"
      422:
        description: "A JSON user object not deserializable, check your JSON."
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "A JSON user object not deserializable, check your JSON."
    definitions:
      User:
        type: object
        properties:
          username:
            type: string
            description: unique username
          gecos:
            type: string
            description: common name and family name
          role:
            type: string
            description: assigned role
          password:
            type: string
          email:
            type: string
          phone:
            type: string
          IM:
            type: string
          isLocked:
            type: boolean
            description: Boolean if the user is locked or not
        required:
          - username
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
    """
    if(request.method == 'POST'): 
        user = None
        # reconstruct user from JSON user object
        try:
            user = User.from_json(request.data)
        except ValueError as e:
            return jsonify({'messg':f'user not deserializable, check your JSON. Caught an exception: {e}'}) , 422
        except KeyError as ke:
            return jsonify({'messg':f'user not deserializable, check your JSON. Caught an exception: {ke}'}) , 422
        if user is None:
            return jsonify({'messg':f'user not deserializable, check your JSON'}) , 422
        # chk if username and user.username are the same
        if username != user.username:
            return jsonify({'messg':f'username and user.username do not match'}) , 422
        # chk if user with that username already exist in users_datastore
        if users_datastore.is_user(username):
            return jsonify({'messg':f'user with this username={user.username} already exist'}) , 409
        # mostly fine, try adding User object to users_datastore
        try:
            my_username = request.user['username'] # request.user is decoded from jwt token
            my_role = users_datastore.get_my_role(my_username) # fetch my role
            users_datastore.add(user, my_role) # save in users_datastore list if role OK
            return jsonify({'messg':f'user username={user.username} saved'}) , 201
        except ValueError as e:
            return jsonify({'messg':f'user not deserializable, check your JSON. Caught an exception: {e}'}) , 422
        except KeyError as ke:
            return jsonify({'messg':f'Important data missing, check your JSON. Caught an exception: {ke}'}) , 422
    else:
        return

# update field(s) of a user if username!=None
#   and new values of fields are passed in the request body
# NOTE: wull NOT change username (which is a key) 
#   or password (which is sensitive and updated via special route)
@bp.route('/<username>', methods = ['PATCH'])
@jwt_required
def user_patch(username=None):
    """Endpoint to update existing user data
    ---
    tags: 
        - User directory
    description: Endpoint to update existing user data
    parameters:
      - name: username
        in: path
        type: string
        required: true
      - name: body
        in: body
        schema:
          $ref: "#/definitions/User"
        required: true
    responses:
      200:
        description: "User object updated"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "User object updated"
      400:
        description: "Invalid or missing JSON payload"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "Invalid or missing JSON payload"
      401:
        description: UNAUTHORIZED
        examples:
          application/json:
            "error": "Missing Authorization header"
      404:
        description: "User to update not found"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "User to update not found"
      422:
        description: "A JSON user object not deserializable, check your JSON."
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "A JSON user object not deserializable, check your JSON."
    definitions:
      User:
        type: object
        properties:
          username:
            type: string
            description: unique username
          gecos:
            type: string
            description: common name and family name
          role:
            type: string
            description: assigned role
          password:
            type: string
          email:
            type: string
          phone:
            type: string
          IM:
            type: string
          isLocked:
            type: boolean
            description: Boolean if the user is locked or not
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
    """
    if(request.method == 'PATCH'): 
        if not username:
            return jsonify({'messg':f'No username provided'}) , 404
        if not users_datastore.is_user(username):
            return jsonify({'messg':f'No such user: {username}'}) , 404
        # get JSON update object
        update_json_object = request.get_json()
        if not update_json_object:
            return jsonify({'messg': 'Invalid or missing JSON payload'}), 400
        
        try:
            my_username = request.user['username'] # request.user is decoded from jwt token
            # patch in users_datastore list if role OK
            users_datastore.update_user_from_json(username, update_json_object, my_username) 
            return jsonify({'messg':f'User username={username} updated'}) , 200
        except ValueError as e:
            return jsonify({'messg':f'User not updated, check your JSON. Error: {e}'}) , 400
        except KeyError as ke:
            return jsonify({'messg':f'Important data missing, check your JSON. Caught an exception: {ke}'}) , 422


# update field(s) of a user if username!=None
#   and new values of fields are passed in the request body
# NOTE: wull NOT change username (which is a key) 
#   or password (which is sensitive and updated via special route)
@bp.route('/<username>/password', methods = ['PATCH'])
@jwt_required
def user_password_patch(username=None):
    """Endpoint to update existing users' password
    ---
    tags: 
        - User directory
    description: Endpoint to update existing users' password
    parameters:
      - name: username
        in: path
        type: string
        required: true
      - name: body
        in: body
        schema:
          $ref: "#/definitions/UserPasswordChange"
        required: true
    responses:
      200:
        description: "User password updated"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "User password updated"
      400:
        description: "Invalid or missing JSON payload"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "Invalid or missing JSON payload"
      401:
        description: UNAUTHORIZED
        examples:
          application/json:
            "error": "Missing Authorization header"
      404:
        description: "User to update not found"
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "User to update not found"
      422:
        description: "A JSON user object not deserializable, check your JSON."
        schema:
          $ref: "#/definitions/ErrMsg"
        examples:
          application/json:
            messg: "A JSON user object not deserializable, check your JSON."
    definitions:
      UserPasswordChange:
        type: object
        properties:
          old_password:
            type: string
            description: old password for the user
          new_password:
            type: string
            description: new password for the user
            required: true
      ErrMsg:
        type: object
        properties:
          messg:
            type: string
    """
    if(request.method == 'PATCH'): 
        if not username:
            return jsonify({'messg':f'No username provided'}) , 404
        if not users_datastore.is_user(username):
            return jsonify({'messg':f'No such user: {username}'}) , 404
        # get JSON update object
        update_json_object = request.get_json()
        if not update_json_object:
            return jsonify({'messg': 'Invalid or missing JSON payload'}), 400
        
        try:
            my_username = request.user['username'] # request.user is decoded from jwt token
            # patch in users_datastore list if role OK
            users_datastore.update_user_password_from_json(username, update_json_object, my_username) 
            # TODO: if username == my_username and execution is here, this means password is changed
            #   so maybe JWT token should be invalidated here?
            # otherwise, same token could be re-used until expiry
            return jsonify({'messg':f'User username={username} updated'}) , 200
        except ValueError as e:
            return jsonify({'messg':f'User not updated, check your JSON. Error: {e}'}) , 400
        except KeyError as ke:
            return jsonify({'messg':f'Important data missing, check your JSON. Caught an exception: {ke}'}) , 422
