import os
import json
from json import JSONDecodeError


roles = ['owner','admin','user']

default_users_list = { 'users' : [
    {'username': 'alice', 'gecos': 'Alice A', 'role': 'admin',
     'password': 'password1', 'email':'alice@example.com',
     'phone':'212-123-4567', 'IM':'alice@jabber.example.com',
     "isLocked": False},
    {'username': 'bob', 'gecos': 'Bob B', 'role': 'owner',
     'password': 'password2', 'email':'bob@example.com',
     'phone':'+49-212-1234567', 'IM':'bob@xmpp.tor.org',
     "isLocked": False},
    {'username': 'charlie','gecos': 'Charlie C', 'role': 'user',
     'password': 'password3', 'email':'charlie@anon.penet.fi', 'phone':'',
     'IM':'bob@chat.well.com', "isLocked": False}
] }

class CustomSavingUsersError(Exception):
    """Custom exception class"""
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

# helper
def is_flat_json(json_obj):
    """
    Checks if the JSON object contains only flat values (no nested JSON or arrays).

    :param json_obj: A dictionary representing a JSON object.
    :return: True if flat, False otherwise.
    """
    if not isinstance(json_obj, dict):
        raise ValueError("Input must be a dictionary representing a JSON object.")

    for key, value in json_obj.items():
        if isinstance(value, (dict, list)):
            return False
    return True

# hash input string for password puroses
def hash_me(string_to_hash):
    import hashlib
    # Create a SHA-512 hash object
    sha512_hash = hashlib.sha512()
    # Update the hash object with the string
    sha512_hash.update(string_to_hash.encode())
    # Get the hexadecimal representation of the hash
    hashed_string = sha512_hash.hexdigest()
    #print("SHA-512 Hashed String:", hashed_string)
    return(hashed_string) 


class User:
    def __init__(self, username, gecos, role, password, email=None, phone=None, IM=None, isLocked=False):
        self.username = username
        self.gecos = gecos
        self.role = role
        self.password = password
        self.email = email
        self.phone = phone
        self.IM = IM
        self.isLocked = isLocked

    def is_authenticated(self):
        return True

    def is_active(self):
        return True

    def is_anonymous(self):
        return False

    def is_admin(self):
        return self.role == 'admin'

    def is_owner(self):
        return self.role == 'owner'

    def get_id(self):
        return self.username

    def to_dict(self):
        return {
            'username': self.username,
            'gecos': self.gecos,
            'role': self.role,
            'password': self.password,
            'email': self.email,
            'phone': self.phone,
            'IM': self.IM,
            'isLocked': self.isLocked
        }

    def to_json(self):
        return json.dumps(self.to_dict(), indent=2)
    
    @staticmethod
    def from_dict(data):
        user = User( data.get('username'), 
                    data.get('gecos'),
                    data.get('role'),
                    data.get('password'),
                    data.get('email'),
                    data.get('phone'),
                    data.get('IM'),
                    data.get('isLocked') )
        return user

    @staticmethod
    def from_json(s):
        data = json.loads(s)
        return User.from_dict(data)


class Users:
    def __init__(self, data=None):
        self.users = []
        if data != None:
            for d in data.get('users'):
                u = User.from_dict(d)
                self.users.append(u)

    @staticmethod
    def from_dict(data):
        users = Users(data)
        return users

    def to_dict(self):
        data = {}
        data['users'] = [u.to_dict() for u in self.users if u is not None]
        return data
    
    def to_json(self):
        return json.dumps(self.to_dict(), indent=2)

    @staticmethod
    def from_json(s):
        data = json.loads(s)
        return Users.from_dict(data)

    @staticmethod
    def load_from_file(filename):
        with open(filename, 'r') as f:
            data = json.load(f)
            return Users.from_dict(data)

    def load(self):
        # load admin data: user_role, user_data from json files
        try:
            # Retrieve JSON data from the file
            file_path = os.path.join('data','users.json')
            with open(file_path, 'r') as f:
                data = json.load(f)
                u = Users.from_dict(data)
                self.users = u.users
        except FileNotFoundError:
            #logging.warrning(f"The file '{file_path}' does not exist.")
            data = default_users_list
            if data != None:
                u = Users.from_dict(data)
                self.users = u.users
            for u in self.users: # hash opentext passwords
                u.password = hash_me(u.password)
            self.save()
        except json.JSONDecodeError as e:
            ''' msg, The unformatted error message.
            doc, The JSON document being parsed.
            pos, The start index of doc where parsing failed.
            lineno, The line corresponding to pos.
            colno, The column corresponding to pos. '''
            # logging.warrning(f"DecodeError '{msg}' in document '{doc}' at position '{pos}'")
            #print(f"DecodeError in document at position ")
        else:
            # logging.info('admin data loaded successfully'):
            print('admin data loaded successfully')

    # save admin data: user_role, user_data to json files    
    def save(self):
        if self.users != []:
            if self.ensure_path_exists('data'):
                with open(os.path.join('data','users.json'), 'w') as f:
                    json.dump(self.to_dict(), f, indent=4)
        else:
            raise CustomSavingUsersError("Saving zero users")

    def ensure_path_exists(self,path):
        """
        Checks if a path exists. If not, creates the directory.

        :param path: The directory path to check and create if necessary.
        :return: True if the path exists or was created successfully, False otherwise.

        path_to_check = "./example_directory"
        ensure_path_exists(path_to_check)
        """
        try:
            if not os.path.exists(path):
                os.makedirs(path)
            return True
        except Exception as e:
            #print(f"Error creating path: {e}")
            return False


    # copy from [{},]
    def copy_from(self,x):
        self.users = x
        
    # return users list of dict
    def get_users(self):
        return self.users

    # check consistensy of user data. every user has to have exactly one role, AND only one is allowed to be the owner
    def chk_user_data(self):
        if self.users in ([], None):
            return True
        # Check if all roles are from the predefined set
        # Initialize variables to keep track of owner count and roles
        consistent = None
        owner_count = 0
        roles_count = {'owner': 0, 'admin': 0, 'user': 0}
        # Iterate through the loaded users
        for user in self.users:
            # the role of the current user has to be in roles
            if user.role not in roles:
                print('username:',user.username,' role:', user.role,' roles:', roles)
                raise Exception('username:',user.username,' role:', user.role,' roles:', roles)
                return False
            # Check if the current user is an owner
            if user.role == 'owner':
                owner_count += 1
        # Check consistency
        if owner_count == 1:
            consistent = True
        else:
            consistent = False
        # Print the result
        if consistent:
            print("User data is consistent.")
        else:
            print("User data is inconsistent.")
            print('users =',self.users)
            print('roles_count =',roles_count)
        return consistent

    # add new user with defined username, password, role, return True if OK
    # TODO: to be phased out, add(User) should be used
    def new_user(self, my_role, username, gecos, password, role):
        # am i allowed?
        if my_role not in roles:
            return False
        if my_role == 'user':
            print('user cannot create new users')
            return False    
        if (my_role == 'admin') and (role!='user'):
            print('admin can only create new user users')
            return False
        # check if role NOT owner - he is the first one and only one
        if role == 'owner':
            print('Cannot add owner, there can be only one, sorry')
            return False
        # than chk if name uniquie
        if self.is_user(username):
            print('username already taken')
            return False
        # than if role OK
        if not (role in  roles):
            print('role not existing')
            return False
        # password exist?
        if password == '':
            print('empty password?')
            return False    
        # so, NOT owner, role OK, username OK, password OK, create new
        #self.users.append({'username': username, 'gecos': gecos, 'role': role, 'password': password, 'isLocked': False}) # one way
        self.users.append( User(username, gecos, role, password, email=None, phone=None, IM=None, isLocked=False) )
        # chk for consistency and save if OK
        if self.chk_user_data():
            self.save()
        else:
            print("chk_user_data consistency failure, admin data not saved")
        return True

    # TODO: to be phased in, instead of new_user
    # add new user object to users
    def add(self, user_object, my_role):
        
        # set of rules that has to be OKed in order to add user object to users
        def add_rules_control(user_object, my_role):
            # am i allowed?
            if my_role not in roles:
                raise ValueError(f'my_role=={my_role} not in known roles')
            if my_role == 'user':
                raise ValueError('role user cannot create new users')
            if (my_role == 'admin') and (user_object.role!='user'):
                raise ValueError('role admin can only create new user users')
            # check if role NOT owner - he is the first one and only one
            if user_object.role == 'owner':
                raise ValueError('Cannot add owner, there can be only one, sorry')
            # than chk if name uniquie
            if self.is_user(user_object.username):
                raise ValueError(f'username {user_object.username} already taken')
            # password exist?
            if user_object.password == '':
                raise ValueError('empty password?')
            return True
        
        # so, NOT owner, role OK, username OK, password OK, create new
        if add_rules_control(user_object, my_role):
            # hash password before saving
            user_object.password = hash_me(user_object.password)
            self.users.append( user_object )
        # chk for consistency and save if OK
        if self.chk_user_data():
            self.save()
        else:
            raise ValueError("chk_user_data consistency failure, data not saved")
        return True



    # delete user with defined username, return True if OK
    # update dicts
    def del_user(self, my_role, username):
        if self.is_user(username) == False:
            # no user w that username
            return False
        if my_role == 'user' :
            # not allowed to delete users
            return False
        self.users.remove( self.get_user(username) )
        # chk for consistency and save if OK
        if self.chk_user_data():
            self.save()
        else:
            print("chk_user_data consistency failure, admin data not saved")
        return True



    # depending on my_role, change role of existing user with defined username, return True if OK
    # update dicts
    def admin_change_user_role(self, my_role, username, new_role):
        # chk if username exist
        # Check if 'alice' exists in the list of users
        username_exists = any(self.user.username == username for user in self.users)
        if not username_exists :
            #print('username not exist')
            return False
        # Find the user_data by username 
        user_data = next((user for user in self.users if user.username == username), None)
        # than if role OK
        # chk if username has a role
        if not (user_data.role in roles):
            print('username does not have a role')
            return False
        # check if role NOT owner - he is the first one and only one
        if new_role == 'owner':
            print('Cannot add owner, there can be only one, sorry')
            return False
        if not (new_role in  roles):
            print('new role not existing')
            return False
        if not (my_role in  roles):
            print('you have no role')
            return False
        # has right to change roles?
        if my_role == 'user':
            print('cannot change roles')
            return False    
        if my_role == 'admin':
            print('cannot elevate roles')
            return False
        # so, only owner can elevate/downgrade user/admin
        # Find the index of the user in the list of users
        index_username = next((index for index, user in enumerate(self.users) if user.username == username), None)
        # Check again if username exists and update her password if found
        if index_username is not None:
            self.users[index_username]['role'] = new_role # Set the new role for username
            #print("Role has been changed.")
        else:
            print("User not found.")
        # chk for consistency and save if OK
        if self.chk_user_data():
            self.save()
        else:
            print("chk_user_data consistency failure, admin data not saved")
        return True

    # depending on my_username get a) my_role return a list of possoble roles I can give to the other user
    # a subset of roles = ['owner','admin','user']
    def possible_roles(self, my_username):
        my_role = self.get_my_role(my_username)
        if my_role == 'user':
            return [] # this should never happen
        if my_role == 'admin':
            return ['user']
        return ['admin','user'] # owner can promote


    def block_user(self, my_role, username):
        '''
        Owner can block/unblock anyone except himself
        Admin can block/unblock users only
        Users cannot block/unblock anyone
        '''
        if my_role not in roles: # if trash role
            return False
        u = self.get_user(username) 
        if u is None: # if nonexisting username
            return False
        if u.role=='owner': 
            return False # owner cannot be blocked
        # if my_role=='user' return False # user has no admin righs
        if my_role=='user': 
            return False # user has no admin righs
        # admin can admin users only
        if (my_role=='admin') and (u.role!='user'): 
            return False # admin can admin users only
        # else means I am owner, I can block anyone
        # block by prefixing password with !
        if u.isLocked:
            return True # already blocked
        else:
            u.isLocked=True
        # if you disable psswd, do it here
        # chk for consistency and save if OK
        if self.chk_user_data():
            self.save()
        else:
            print("chk_user_data consistency failure, admin data not saved")
        return True

    # unblock user, can be done by owner for all, by admin for all users
    def unblock_user(self, my_role, username):
        u = self.get_user(username)
        if u == None:
            #print("No User found.")
            return False
        # if my_role=='user' return False # user has no admin righs
        if my_role=='user': 
            return False # user has no admin righs
        # admin can admin users only
        if (my_role=='admin') and (u.role != 'user'): 
            return False # admin can unblock users only
        # else means I am owner, I can block anyone
        # unblock by removing ! as password prefix if there
        if u.isLocked:
            u.isLocked = False
            #print("Password has been unblocked.")
            # chk for consistency and save if OK
            if self.chk_user_data():
                self.save()
            else:
                print("chk_user_data consistency failure, admin data not saved")
            return True


    # is user blocked
    def is_block_user(self, username):
        u = self.get_user(username)
        if u != None:
            return u.isLocked
        return False

    # is user
    def is_user(self, username):
        u = self.get_user(username)
        return(u != None)

    # get user data 
    # Find the user with username 'bob' or None
    def get_user(self, username):
        result = [user for user in self.users if user.username == username]
        if result == []:
            return None
        return result[0]

    # get my role based on my username
    def get_my_role(self, my_username):
        u = self.get_user(my_username)
        if u is not None:
            return u.role
        return None

    # is this password for this username in userdata
    def is_password(self, username,password):
        u = self.get_user(username)
        if u != None:
            return (password == u.password)
        return False

    # change password from old to new for this username in userdata
    def change_password(self, username,old_password, new_password):
        u = self.get_user(username)
        if u == None:
            return False # no such user
        if(old_password != u.password):
            return False # didn know old password
        u.password = new_password
        # chk for consistency and save if OK
        if self.chk_user_data():
            self.save()
        else:
            print("chk_user_data consistency failure, admin data not saved")
        return True

    # search for specific users by search_term (only ascii, no special chars, one word only) 
    # return all matching substring by 'uid', 'gecos', 'cn', 'sn' and 'mail' attributes
    def search_users(self, search_term, exclude_username):
        result = []
        tmp_users = [user for user in self.users if user.username != exclude_username]
        for u in tmp_users:
            if (search_term in u.username) or (search_term in u.gecos):
                result.append(u)
        return result

    # update user in users
    # argument user_object_dict might have a subset of attributes but MUST contain username. missing keys will be blank
    # any xtra key:value par will be ignored
    # returns boolean success
    def update_user_old(self, user_object_dict):
        success  = False
        # Find the index of the user in the list of users
        index_username = next((index for index, user in enumerate(self.users) if user.username == user_object_dict['username']), None)
        if index_username is not None: # bingo we found the user
            for key in user_object_dict:
                if (key not in ['submit', 'csrf_token']):
                    # Check if the key/value even exist
                    if key in self.users[index_username]:
                        # Check if the value is different
                        if self.users[index_username][key] != user_object_dict[key]:
                            # Update the value
                            self.users[index_username][key] = user_object_dict[key]
                            #print(f"Updated {key} for user to '{user_object_dict[key]}'.")
                            success  = True
                        else:
                            print(f"No change for key '{key}'.")
                    else: #expand self.users[index_username] with new key:value pair
                        self.users[index_username][key] = user_object_dict[key]
                        #print(f"Expanded {key} for user to '{user_object_dict[key]}'.")
                        success  = True
        else:
            #print("User not found.")
            success  = False
        if success:
            if not self.chk_user_data():
                success = False
        return success

    # update user in users
    # argument user_object_dict might have a subset of attributes but MUST contain username. missing keys will be blank
    # any xtra key:value par will be ignored
    # returns boolean success
    def update_user(self, user_object_dict):
        success = False
        u = user_object_dict.get('username')
        if u in ('', None):
            return False
        target_user = self.get_user(u)
        if target_user in ('', None):
            return False
        # build target_user from user_object_dict
        updated_user = User.from_dict(user_object_dict)
        if updated_user.username != target_user.username:
            return False # they have different usernames keys
        # Update logic: find and replace
        for i, u in enumerate(self.users):
            if u.username == updated_user.username:  # Matching criterion 
                updated_user.password = u.password # save old password
                self.users[i] = updated_user
                success = True
                break
        if success:
            if not self.chk_user_data():
                success = False
        return success

    # patch in users_datastore list if role OK
    def update_user_from_json(self, username, update_json_object, my_username):

        def update_user_attributes(user, update_json_object, skip):
            """
            Updates the attributes of a User object based on a JSON object,
            considering the skip list.
            
            :param user: The User object to update.
            :param update_json_object: A dictionary with the updates.
            :param skip: A list of attributes to skip during the update.
            """
            for key, value in update_json_object.items():
                if hasattr(user, key) and key not in skip:
                    setattr(user, key, value)    
        
        # chk for consistency of arguments, and raise exceptions if not OK
        if not self.is_user(username):
            raise ValueError(f'user {username} for update not found')
        # update_json_object exist?
        if update_json_object is None:
            raise ValueError(f'update JSON object not found')
        # is update_json_object plain structure?
        if not is_flat_json(update_json_object):
            raise ValueError(f'incorrect update structure, check your JSON')
        # my_username exist?
        if not my_username:
            raise ValueError(f'username trying to perform update does not exist')
        # my_username exist in data store?
        if not self.is_user(my_username):
            raise ValueError(f'user {my_username} trying to perform update not found')
        # there are 2 use cases: a) updating self data, and b) updating others' data
        u = self.get_user(username)
        if username == my_username: # updating self
            skip = ['username','password','role'] # cannot change this
            update_user_attributes(u, update_json_object, skip)# update 
        else: # updating others
            my_role = self.get_my_role(my_username) # fetch my role
            if my_role == 'user':
                raise ValueError("role 'user' cannot update others") 
            # 'admin' or 'owner' can update others
            skip = ['username','password'] # cannot change this
            update_user_attributes(u, update_json_object, skip)# update         
        self.save()



    # update users' password in users
    # argument user_passwordobject_dict 
    # patch in users_datastore list if role OK
    def update_user_password_from_json(self, username, update_json_object, my_username):

        # chk for consistency of arguments, and raise exceptions if not OK
        if not self.is_user(username):
            raise ValueError(f'user {username} for password update not found')
        # update_json_object exist?
        if update_json_object is None:
            raise ValueError(f'password update JSON object not found')
        # is update_json_object plain structure?
        if not is_flat_json(update_json_object):
            raise ValueError(f'incorrect password update structure, check your JSON')
        # my_username exist?
        if not my_username:
            raise ValueError(f'username trying to perform update does not exist')
        # my_username exist in data store?
        if not self.is_user(my_username):
            raise ValueError(f'user {my_username} trying to perform password update not found')
        # there are 2 use cases: 
        #   a) updating self password data (has to supply old_password as well), and 
        #   b) updating others password data
        u = self.get_user(username)
        if username == my_username: # updating self, role does not mater
            if ((not update_json_object.get('old_password')) or
                (not update_json_object.get('new_password'))):
                raise ValueError(f'old or new password missing. password not changed')

            old_password = hash_me(update_json_object.get('old_password'))
            new_password = hash_me(update_json_object.get('new_password'))
            if u.password == old_password: # he knows the password, OK
                u.password = new_password
            else:
                raise ValueError(f'Incorrect old password value. password not changed')

        else: # updating others, role does mater, old password does not matter
            my_role = self.get_my_role(my_username) # fetch my role
            if my_role == 'user':
                raise ValueError("role 'user' cannot update paswword of others") 
            if my_role == 'admin' and u.role=='owner':
                raise ValueError("role 'admin' cannot update password of 'owner'") 
            # 'admin' or 'owner' can update others
            if not update_json_object.get('new_password'):
                raise ValueError(f'password missing. password not changed')
            new_password = hash_me(update_json_object.get('new_password'))
            u.password = new_password
            
        self.save()

