import pytest
import requests

BASE_URL = "http://localhost:5000/auth"
BASE_URL_USERS = "http://localhost:5000/users"



@pytest.fixture
def setup_tree():
    # Step 1: Generate a token for bob
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "bob", "password": "password2"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "token" in token_data, "Token not found in response"
    token = token_data["token"]
    assert token.startswith("eyJ"), "Token does not appear to be a valid JWT"
    return token


def test_get_token(setup_tree):
    token = setup_tree
    assert token.startswith("eyJ"), "Token does not appear to be a valid JWT"

def test_token_validation(setup_tree):
    token = setup_tree

    # Step 2: Validate the token
    test_response = requests.get(
        f"{BASE_URL}/test",
        headers={"Authorization": f"Bearer {token}"}
    )
    # Assert token validation is successful
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code}"
    test_data = test_response.json()
    assert "message" in test_data, "Message not found in response"
    assert test_data["message"] == "Access granted for user bob.", "Unexpected validation message"




# list all users
def test_list_users(setup_tree):
    token = setup_tree
    test_response = requests.get(
        f"{BASE_URL_USERS}",
        headers={"Authorization": f"Bearer {token}"}
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code}"
    test_data = test_response.json()
    assert 'charlie' in test_data, "Message not found in response"

# search users by useranme substring - user found
def test_search_users_found(setup_tree):
    token = setup_tree
    test_response = requests.get(
        f"{BASE_URL_USERS}?username=bob",
        headers={"Authorization": f"Bearer {token}"}
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code}"
    test_data = test_response.json()
    assert 'bob' in test_data, "Message not found in response"

# search users by useranme substring - user NOT found
def test_search_users_not_found(setup_tree):
    token = setup_tree
    test_response = requests.get(
        f"{BASE_URL_USERS}?username=mary",
        headers={"Authorization": f"Bearer {token}"}
    )
    assert test_response.status_code == 404, f"Unexpected status code: {test_response.status_code}"
    test_data = test_response.json()
    assert test_data == {'messg': 'target entity mary not found'}


# get user data by username, found
def test_users_get(setup_tree):
    token = setup_tree
    test_response = requests.get(
        f"{BASE_URL_USERS}/bob",
        headers={"Authorization": f"Bearer {token}"}
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code}"
    test_data = test_response.json()
    assert test_data == {'username': 'bob', 'gecos': 'Bob B', 'role': 'owner', 'password': '*****', 'email': 'bob@example.com', 'phone': '+49-212-1234567', 'IM': 'bob@xmpp.tor.org', 'isLocked': False}

# get user data by username, NOT found
def test_users_get_wrong(setup_tree):
    token = setup_tree
    test_response = requests.get(
        f"{BASE_URL_USERS}/boba",
        headers={"Authorization": f"Bearer {token}"}
    )
    assert test_response.status_code == 404, f"Unexpected status code: {test_response.status_code}"
    test_data = test_response.json()
    assert 'messg' in test_data, "Message not found in response"
    assert test_data == {'messg': 'user with this username=boba not found'}


# add new user 
def test_users_add(setup_tree):
    token = setup_tree
    test_response = requests.post(
        f"{BASE_URL_USERS}/edi",
        headers={"Authorization": f"Bearer {token}"},
        json={
                    "username": "edi",
                    "gecos": "Eddie Doe",
                    "role": "admin",
                    "password": "securepassword123",
                    "email": "eddoe@example.com",
                    "phone": "+1234567890",
                    "IM": "eddoe123",
                    "isLocked": False
                  }
    )
    assert test_response.status_code == 201, f"Unexpected status code: {test_response.status_code}"
    test_data = test_response.json()
    assert test_data == {'messg': 'user username=edi saved'}


# delete user 
def test_user_delete(setup_tree):
    token = setup_tree
    test_response = requests.delete(
        f"{BASE_URL_USERS}/edi",
        headers={"Authorization": f"Bearer {token}"},
    )
    assert test_response.status_code == 204, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

# delete self
def test_user_delete_self(setup_tree):
    token = setup_tree
    test_response = requests.delete(
        f"{BASE_URL_USERS}/bob",
        headers={"Authorization": f"Bearer {token}"},
    )
    assert test_response.status_code == 400, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

# update user
def test_user_update(setup_tree):
    token = setup_tree
    test_response = requests.get(
        f"{BASE_URL_USERS}/alice",
        headers={"Authorization": f"Bearer {token}"},
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code} {test_response.json()}"
    test_data = test_response.json()
    assert  test_data == {'IM': 'alice@jabber.example.com',
          'email': 'alice@example.com',
          'gecos': 'Alice A',
          'isLocked': False,
          'password': '*****',
          'phone': '212-123-4567',
          'role': 'admin',
          'username': 'alice'}

    test_response = requests.patch(
        f"{BASE_URL_USERS}/alice",
        headers={"Authorization": f"Bearer {token}"},
        json = {'gecos': 'Alice B',
          'isLocked': True,
          'password': '**',
          'phone': '063',
          'role': 'user',
          'username': '1alice'}
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code} {test_response.json()}"


    test_response = requests.get(
        f"{BASE_URL_USERS}/alice",
        headers={"Authorization": f"Bearer {token}"},
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code} {test_response.json()}"
    test_data = test_response.json()
    assert  test_data == {'IM': 'alice@jabber.example.com',
          'email': 'alice@example.com',
          'gecos': 'Alice B',
          'isLocked': True,
          'password': '*****',
          'phone': '063',
          'role': 'user',
          'username': 'alice'}

    test_response = requests.patch(
        f"{BASE_URL_USERS}/alice",
        headers={"Authorization": f"Bearer {token}"},
        json = {'gecos': 'Alice A',
          'isLocked': False,
          'password': '*****',
          'phone': '212-123-4567',
          'role': 'admin',
          'username': 'alice'}
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    test_response = requests.get(
        f"{BASE_URL_USERS}/alice",
        headers={"Authorization": f"Bearer {token}"},
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code} {test_response.json()}"
    test_data = test_response.json()
    assert  test_data == {'IM': 'alice@jabber.example.com',
          'email': 'alice@example.com',
          'gecos': 'Alice A',
          'isLocked': False,
          'password': '*****',
          'phone': '212-123-4567',
          'role': 'admin',
          'username': 'alice'}

# password change, self
def test_self_password_change(setup_tree):
    # I am charlie, wanna change my password
    # Step 1: Generate a token for charlie
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "charlie", "password": "password3"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "token" in token_data, "Token not found in response"
    token = token_data["token"]
    assert token.startswith("eyJ"), "Token does not appear to be a valid JWT"


    # this should NOT work, no old password
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'new_password': 'password4',
          'junk': '063'}
    )
    assert test_response.status_code == 400, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    # this should NOT work, no new password
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'old_password': 'password3',
          'junk': '063'}
    )
    assert test_response.status_code == 400, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    # this should NOT work, no JSON 
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie/password",
        headers={"Authorization": f"Bearer {token}"},
    )
    assert test_response.status_code == 400, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    # this should NOT work, user try changing admins' password
    test_response = requests.patch(
        f"{BASE_URL_USERS}/alice/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'old_password': 'password1',
          'new_password': 'password4'}
    )
    assert test_response.status_code == 400, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    # this should NOT work, user try changing owners' password
    test_response = requests.patch(
        f"{BASE_URL_USERS}/bob/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'old_password': 'password2',
          'new_password': 'password4'}
    )
    assert test_response.status_code == 400, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    
    # this should work, password change
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'old_password': 'password3',
          'new_password': 'password4',
          'junk': '063'}
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    # now password should be changed
    # test login with old token wich is valied until change or token expiry
    # this should work
    token_response = requests.get(
        f"{BASE_URL}/test",
        headers={"Authorization": f"Bearer {token}"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()

    # login charlie again with old credentials, should fail
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "charlie", "password": "password3"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 401, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "error" in token_data, "Token not found in response"

    # login charlie again with new credentials, should work
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "charlie", "password": "password4"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "token" in token_data, "Token not found in response"
    token = token_data["token"]
    assert token.startswith("eyJ"), "Token does not appear to be a valid JWT"

    # now password should be changed, and token is changed
    # test login with new password, and token
    token_response = requests.get(
        f"{BASE_URL}/test",
        headers={"Authorization": f"Bearer {token}"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()


    # this should work, return to old password
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'old_password': 'password4',
          'new_password': 'password3',
          'junk': '063'}
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code} {test_response.json()}"


    # login charlie again with original credentials and get latest token, should work
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "charlie", "password": "password3"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "token" in token_data, "Token not found in response"
    token = token_data["token"]
    assert token.startswith("eyJ"), "Token does not appear to be a valid JWT"


    # and test original password
    # now password should be changed
    # test login with new password, 
    token_response = requests.get(
        f"{BASE_URL}/test",
        headers={"Authorization": f"Bearer {token}"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()


# password change, others
def test_others_password_change(setup_tree):
    # I am alice (admin), wanna change password for charlie
    # Step 1: Generate a token for alice
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "alice", "password": "password1"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "token" in token_data, "Token not found in response"
    token = token_data["token"]
    assert token.startswith("eyJ"), "Token does not appear to be a valid JWT"


    # this should NOT work, wrong username
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie1/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'new_password': 'password4',
          'junk': '063'}
    )
    assert test_response.status_code == 404, f"Unexpected status code: {test_response.status_code} {test_response.json()}"


    # this should NOT work, no new password
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'old_password': 'password3',
          'junk': '063'}
    )
    assert test_response.status_code == 400, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    # this should NOT work, no JSON 
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie/password",
        headers={"Authorization": f"Bearer {token}"},
    )
    assert test_response.status_code == 400, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    # this should NOT work, admin try changing owners' password
    test_response = requests.patch(
        f"{BASE_URL_USERS}/bob/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'new_password': 'password22'}
    )
    assert test_response.status_code == 400, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    # this should work, password change
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'new_password': 'password4',
          'junk': '063'}
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code} {test_response.json()}"

    # login as charlie with old credentials, should fail
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "charlie", "password": "password3"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 401, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "error" in token_data, "Token not found in response"

    # login charlie again with new credentials, should work
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "charlie", "password": "password4"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "token" in token_data, "Token not found in response"
    token = token_data["token"]
    assert token.startswith("eyJ"), "Token does not appear to be a valid JWT"

    # now password should be changed, and token is changed
    # test login with new password, and token
    token_response = requests.get(
        f"{BASE_URL}/test",
        headers={"Authorization": f"Bearer {token}"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()

    # login as bob the owner
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "bob", "password": "password2"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "token" in token_data, "Token not found in response"
    token = token_data["token"]
    assert token.startswith("eyJ"), "Token does not appear to be a valid JWT"


    # this should work, return charlie to old password
    test_response = requests.patch(
        f"{BASE_URL_USERS}/charlie/password",
        headers={"Authorization": f"Bearer {token}"},
        json = {'new_password': 'password3'}
    )
    assert test_response.status_code == 200, f"Unexpected status code: {test_response.status_code} {test_response.json()}"


    # login charlie again with original credentials and get latest token, should work
    token_response = requests.post(
        f"{BASE_URL}/token",
        headers={"Content-Type": "application/json"},
        json={"username": "charlie", "password": "password3"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
    assert "token" in token_data, "Token not found in response"
    token = token_data["token"]
    assert token.startswith("eyJ"), "Token does not appear to be a valid JWT"


    # and test original password
    # now password should be changed
    # test login with new password, 
    token_response = requests.get(
        f"{BASE_URL}/test",
        headers={"Authorization": f"Bearer {token}"}
    )
    # Assert token generation is successful
    assert token_response.status_code == 200, f"Unexpected status code: {token_response.status_code}"
    token_data = token_response.json()
        
